@@ -3,12 +3,12 @@
  * analyze.c
  *	  the postgres statistics generator
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/commands/analyze.c,v 1.25.2.1 2002/02/18 16:04:21 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/commands/analyze.c,v 1.64 2003/10/18 15:38:06 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -18,18 +18,22 @@
 
 #include "access/heapam.h"
 #include "access/tuptoaster.h"
+#include "catalog/catalog.h"
 #include "catalog/catname.h"
 #include "catalog/indexing.h"
+#include "catalog/namespace.h"
 #include "catalog/pg_operator.h"
 #include "catalog/pg_statistic.h"
 #include "catalog/pg_type.h"
 #include "commands/vacuum.h"
 #include "miscadmin.h"
 #include "parser/parse_oper.h"
+#include "parser/parse_relation.h"
 #include "utils/acl.h"
 #include "utils/builtins.h"
 #include "utils/datum.h"
 #include "utils/fmgroids.h"
+#include "utils/lsyscache.h"
 #include "utils/syscache.h"
 #include "utils/tuplesort.h"
 
@@ -53,11 +57,11 @@ typedef enum
  * ignoring wide values will not affect our estimates of histogram bin
  * boundaries very much.
  */
-#define WIDTH_THRESHOLD  256
+#define WIDTH_THRESHOLD  1024
 
 /*
  * We build one of these structs for each attribute (column) that is to be
- * analyzed.  The struct and subsidiary data are in TransactionCommandContext,
+ * analyzed.  The struct and subsidiary data are in anl_context,
  * so they live until the end of the ANALYZE operation.
  */
 typedef struct
@@ -106,7 +110,13 @@ typedef struct
 #define swapDatum(a,b)	do {Datum _tmp; _tmp=a; a=b; b=_tmp;} while(0)
 
 
-static int	MESSAGE_LEVEL;
+/* Default statistics target (GUC parameter) */
+int			default_statistics_target = 10;
+
+
+static int	elevel = -1;
+
+static MemoryContext anl_context = NULL;
 
 /* context information for compare_scalars() */
 static FmgrInfo *datumCmpFn;
@@ -139,7 +149,6 @@ void
 analyze_rel(Oid relid, VacuumStmt *vacstmt)
 {
 	Relation	onerel;
-	Form_pg_attribute *attr;
 	int			attr_cnt,
 				tcnt,
 				i;
@@ -148,25 +157,22 @@ analyze_rel(Oid relid, VacuumStmt *vacstmt)
 				numrows;
 	double		totalrows;
 	HeapTuple  *rows;
-	HeapTuple	tuple;
 
 	if (vacstmt->verbose)
-		MESSAGE_LEVEL = NOTICE;
+		elevel = INFO;
 	else
-		MESSAGE_LEVEL = DEBUG;
+		elevel = DEBUG2;
 
 	/*
-	 * Begin a transaction for analyzing this relation.
-	 *
-	 * Note: All memory allocated during ANALYZE will live in
-	 * TransactionCommandContext or a subcontext thereof, so it will all
-	 * be released by transaction commit at the end of this routine.
+	 * Use the current context for storing analysis info.  vacuum.c
+	 * ensures that this context will be cleared when I return, thus
+	 * releasing the memory allocated here.
 	 */
-	StartTransactionCommand();
+	anl_context = CurrentMemoryContext;
 
 	/*
 	 * Check for user-requested abort.	Note we want this to be inside a
-	 * transaction, so xact.c doesn't issue useless NOTICE.
+	 * transaction, so xact.c doesn't issue useless WARNING.
 	 */
 	CHECK_FOR_INTERRUPTS();
 
@@ -174,56 +180,77 @@ analyze_rel(Oid relid, VacuumStmt *vacstmt)
 	 * Race condition -- if the pg_class tuple has gone away since the
 	 * last time we saw it, we don't need to process it.
 	 */
-	tuple = SearchSysCache(RELOID,
-						   ObjectIdGetDatum(relid),
-						   0, 0, 0);
-	if (!HeapTupleIsValid(tuple))
+	if (!SearchSysCacheExists(RELOID,
+							  ObjectIdGetDatum(relid),
+							  0, 0, 0))
+		return;
+
+	/*
+	 * Open the class, getting only a read lock on it, and check
+	 * permissions. Permissions check should match vacuum's check!
+	 */
+	onerel = relation_open(relid, AccessShareLock);
+
+	if (!(pg_class_ownercheck(RelationGetRelid(onerel), GetUserId()) ||
+		  (pg_database_ownercheck(MyDatabaseId, GetUserId()) && !onerel->rd_rel->relisshared)))
 	{
-		CommitTransactionCommand();
+		/* No need for a WARNING if we already complained during VACUUM */
+		if (!vacstmt->vacuum)
+			ereport(WARNING,
+					(errmsg("skipping \"%s\" --- only table or database owner can analyze it",
+							RelationGetRelationName(onerel))));
+		relation_close(onerel, AccessShareLock);
+		return;
+	}
+
+	/*
+	 * Check that it's a plain table; we used to do this in getrels() but
+	 * seems safer to check after we've locked the relation.
+	 */
+	if (onerel->rd_rel->relkind != RELKIND_RELATION)
+	{
+		/* No need for a WARNING if we already complained during VACUUM */
+		if (!vacstmt->vacuum)
+			ereport(WARNING,
+					(errmsg("skipping \"%s\" --- cannot analyze indexes, views, or special system tables",
+							RelationGetRelationName(onerel))));
+		relation_close(onerel, AccessShareLock);
+		return;
+	}
+
+	/*
+	 * Silently ignore tables that are temp tables of other backends ---
+	 * trying to analyze these is rather pointless, since their contents
+	 * are probably not up-to-date on disk.  (We don't throw a warning
+	 * here; it would just lead to chatter during a database-wide
+	 * ANALYZE.)
+	 */
+	if (isOtherTempNamespace(RelationGetNamespace(onerel)))
+	{
+		relation_close(onerel, AccessShareLock);
 		return;
 	}
 
 	/*
 	 * We can ANALYZE any table except pg_statistic. See update_attstats
 	 */
-	if (strcmp(NameStr(((Form_pg_class) GETSTRUCT(tuple))->relname),
-			   StatisticRelationName) == 0)
+	if (IsSystemNamespace(RelationGetNamespace(onerel)) &&
+	 strcmp(RelationGetRelationName(onerel), StatisticRelationName) == 0)
 	{
-		ReleaseSysCache(tuple);
-		CommitTransactionCommand();
-		return;
-	}
-	ReleaseSysCache(tuple);
-
-	/*
-	 * Open the class, getting only a read lock on it, and check
-	 * permissions. Permissions check should match vacuum's check!
-	 */
-	onerel = heap_open(relid, AccessShareLock);
-
-	if (!(pg_ownercheck(GetUserId(), RelationGetRelationName(onerel),
-						RELNAME) ||
-		  (is_dbadmin(MyDatabaseId) && !onerel->rd_rel->relisshared)))
-	{
-		/* No need for a notice if we already complained during VACUUM */
-		if (!vacstmt->vacuum)
-			elog(NOTICE, "Skipping \"%s\" --- only table or database owner can ANALYZE it",
-				 RelationGetRelationName(onerel));
-		heap_close(onerel, NoLock);
-		CommitTransactionCommand();
+		relation_close(onerel, AccessShareLock);
 		return;
 	}
 
-	elog(MESSAGE_LEVEL, "Analyzing %s", RelationGetRelationName(onerel));
+	ereport(elevel,
+			(errmsg("analyzing \"%s.%s\"",
+					get_namespace_name(RelationGetNamespace(onerel)),
+					RelationGetRelationName(onerel))));
 
 	/*
 	 * Determine which columns to analyze
 	 *
 	 * Note that system attributes are never analyzed.
 	 */
-	attr = onerel->rd_att->attrs;
-	attr_cnt = onerel->rd_att->natts;
-
 	if (vacstmt->va_cols != NIL)
 	{
 		List	   *le;
@@ -235,15 +262,8 @@ analyze_rel(Oid relid, VacuumStmt *vacstmt)
 		{
 			char	   *col = strVal(lfirst(le));
 
-			for (i = 0; i < attr_cnt; i++)
-			{
-				if (namestrcmp(&(attr[i]->attname), col) == 0)
-					break;
-			}
-			if (i >= attr_cnt)
-				elog(ERROR, "ANALYZE: there is no attribute %s in %s",
-					 col, RelationGetRelationName(onerel));
-			vacattrstats[tcnt] = examine_attribute(onerel, i + 1);
+			i = attnameAttNum(onerel, col, false);
+			vacattrstats[tcnt] = examine_attribute(onerel, i);
 			if (vacattrstats[tcnt] != NULL)
 				tcnt++;
 		}
@@ -251,12 +271,14 @@ analyze_rel(Oid relid, VacuumStmt *vacstmt)
 	}
 	else
 	{
-		vacattrstats = (VacAttrStats **) palloc(attr_cnt *
+		attr_cnt = onerel->rd_att->natts;
+		/* +1 here is just to avoid palloc(0) with zero-column table */
+		vacattrstats = (VacAttrStats **) palloc((attr_cnt + 1) *
 												sizeof(VacAttrStats *));
 		tcnt = 0;
-		for (i = 0; i < attr_cnt; i++)
+		for (i = 1; i <= attr_cnt; i++)
 		{
-			vacattrstats[tcnt] = examine_attribute(onerel, i + 1);
+			vacattrstats[tcnt] = examine_attribute(onerel, i);
 			if (vacattrstats[tcnt] != NULL)
 				tcnt++;
 		}
@@ -268,8 +290,7 @@ analyze_rel(Oid relid, VacuumStmt *vacstmt)
 	 */
 	if (attr_cnt <= 0)
 	{
-		heap_close(onerel, NoLock);
-		CommitTransactionCommand();
+		relation_close(onerel, AccessShareLock);
 		return;
 	}
 
@@ -308,14 +329,14 @@ analyze_rel(Oid relid, VacuumStmt *vacstmt)
 	 * Compute the statistics.	Temporary results during the calculations
 	 * for each column are stored in a child context.  The calc routines
 	 * are responsible to make sure that whatever they store into the
-	 * VacAttrStats structure is allocated in TransactionCommandContext.
+	 * VacAttrStats structure is allocated in anl_context.
 	 */
 	if (numrows > 0)
 	{
 		MemoryContext col_context,
 					old_context;
 
-		col_context = AllocSetContextCreate(CurrentMemoryContext,
+		col_context = AllocSetContextCreate(anl_context,
 											"Analyze Column",
 											ALLOCSET_DEFAULT_MINSIZE,
 											ALLOCSET_DEFAULT_INITSIZE,
@@ -355,10 +376,7 @@ analyze_rel(Oid relid, VacuumStmt *vacstmt)
 	 * before we commit.  (If someone did, they'd fail to clean up the
 	 * entries we made in pg_statistic.)
 	 */
-	heap_close(onerel, NoLock);
-
-	/* Commit and release working memory */
-	CommitTransactionCommand();
+	relation_close(onerel, NoLock);
 }
 
 /*
@@ -372,30 +390,26 @@ examine_attribute(Relation onerel, int attnum)
 {
 	Form_pg_attribute attr = onerel->rd_att->attrs[attnum - 1];
 	Operator	func_operator;
-	Oid			oprrest;
 	HeapTuple	typtuple;
 	Oid			eqopr = InvalidOid;
 	Oid			eqfunc = InvalidOid;
 	Oid			ltopr = InvalidOid;
 	VacAttrStats *stats;
 
+	/* Don't analyze dropped columns */
+	if (attr->attisdropped)
+		return NULL;
+
 	/* Don't analyze column if user has specified not to */
-	if (attr->attstattarget <= 0)
+	if (attr->attstattarget == 0)
 		return NULL;
 
 	/* If column has no "=" operator, we can't do much of anything */
-	func_operator = compatible_oper("=",
-									attr->atttypid,
-									attr->atttypid,
-									true);
+	func_operator = equality_oper(attr->atttypid, true);
 	if (func_operator != NULL)
 	{
-		oprrest = ((Form_pg_operator) GETSTRUCT(func_operator))->oprrest;
-		if (oprrest == F_EQSEL)
-		{
-			eqopr = oprid(func_operator);
-			eqfunc = oprfuncid(func_operator);
-		}
+		eqopr = oprid(func_operator);
+		eqfunc = oprfuncid(func_operator);
 		ReleaseSysCache(func_operator);
 	}
 	if (!OidIsValid(eqfunc))
@@ -405,8 +419,7 @@ examine_attribute(Relation onerel, int attnum)
 	 * If we have "=" then we're at least able to do the minimal
 	 * algorithm, so start filling in a VacAttrStats struct.
 	 */
-	stats = (VacAttrStats *) palloc(sizeof(VacAttrStats));
-	MemSet(stats, 0, sizeof(VacAttrStats));
+	stats = (VacAttrStats *) palloc0(sizeof(VacAttrStats));
 	stats->attnum = attnum;
 	stats->attr = (Form_pg_attribute) palloc(ATTRIBUTE_TUPLE_SIZE);
 	memcpy(stats->attr, attr, ATTRIBUTE_TUPLE_SIZE);
@@ -414,23 +427,22 @@ examine_attribute(Relation onerel, int attnum)
 							  ObjectIdGetDatum(attr->atttypid),
 							  0, 0, 0);
 	if (!HeapTupleIsValid(typtuple))
-		elog(ERROR, "cache lookup of type %u failed", attr->atttypid);
+		elog(ERROR, "cache lookup failed for type %u", attr->atttypid);
 	stats->attrtype = (Form_pg_type) palloc(sizeof(FormData_pg_type));
 	memcpy(stats->attrtype, GETSTRUCT(typtuple), sizeof(FormData_pg_type));
 	ReleaseSysCache(typtuple);
 	stats->eqopr = eqopr;
 	stats->eqfunc = eqfunc;
 
+	/* If the attstattarget column is negative, use the default value */
+	if (stats->attr->attstattarget < 0)
+		stats->attr->attstattarget = default_statistics_target;
+
 	/* Is there a "<" operator with suitable semantics? */
-	func_operator = compatible_oper("<",
-									attr->atttypid,
-									attr->atttypid,
-									true);
+	func_operator = ordering_oper(attr->atttypid, true);
 	if (func_operator != NULL)
 	{
-		oprrest = ((Form_pg_operator) GETSTRUCT(func_operator))->oprrest;
-		if (oprrest == F_SCALARLTSEL)
-			ltopr = oprid(func_operator);
+		ltopr = oprid(func_operator);
 		ReleaseSysCache(func_operator);
 	}
 	stats->ltopr = ltopr;
@@ -462,14 +474,14 @@ examine_attribute(Relation onerel, int attnum)
 		 * know it at this point.
 		 *--------------------
 		 */
-		stats->minrows = 300 * attr->attstattarget;
+		stats->minrows = 300 * stats->attr->attstattarget;
 	}
 	else
 	{
 		/* Can't do much but the minimal stuff */
 		stats->algcode = ALG_MINIMAL;
 		/* Might as well use the same minrows as above */
-		stats->minrows = 300 * attr->attstattarget;
+		stats->minrows = 300 * stats->attr->attstattarget;
 	}
 
 	return stats;
@@ -510,8 +522,8 @@ acquire_sample_rows(Relation onerel, HeapTuple *rows, int targrows,
 	/*
 	 * Do a simple linear scan until we reach the target number of rows.
 	 */
-	scan = heap_beginscan(onerel, false, SnapshotNow, 0, NULL);
-	while (HeapTupleIsValid(tuple = heap_getnext(scan, 0)))
+	scan = heap_beginscan(onerel, SnapshotNow, 0, NULL);
+	while ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 	{
 		rows[numrows++] = heap_copytuple(tuple);
 		if (numrows >= targrows)
@@ -527,6 +539,12 @@ acquire_sample_rows(Relation onerel, HeapTuple *rows, int targrows,
 	if (!HeapTupleIsValid(tuple))
 	{
 		*totalrows = (double) numrows;
+
+		ereport(elevel,
+				(errmsg("\"%s\": %u pages, %d rows sampled, %.0f estimated total rows",
+						RelationGetRelationName(onerel),
+						onerel->rd_nblocks, numrows, *totalrows)));
+
 		return numrows;
 	}
 
@@ -620,8 +638,7 @@ pageloop:;
 		 */
 		targbuffer = ReadBuffer(onerel, targblock);
 		if (!BufferIsValid(targbuffer))
-			elog(ERROR, "acquire_sample_rows: ReadBuffer(%s,%u) failed",
-				 RelationGetRelationName(onerel), targblock);
+			elog(ERROR, "ReadBuffer failed");
 		LockBuffer(targbuffer, BUFFER_LOCK_SHARE);
 		targpage = BufferGetPage(targbuffer);
 		maxoffset = PageGetMaxOffsetNumber(targpage);
@@ -641,8 +658,8 @@ pageloop:;
 				goto pageloop;
 			}
 			ItemPointerSet(&targtuple.t_self, targblock, targoffset);
-			heap_fetch(onerel, SnapshotNow, &targtuple, &tupbuffer, NULL);
-			if (targtuple.t_data != NULL)
+			if (heap_fetch(onerel, SnapshotNow, &targtuple, &tupbuffer,
+						   false, NULL))
 			{
 				/*
 				 * Found a suitable tuple, so save it, replacing one old
@@ -676,6 +693,14 @@ pageloop:;
 	 */
 	*totalrows = floor((double) onerel->rd_nblocks * tuplesperpage + 0.5);
 
+	/*
+	 * Emit some interesting relation info 
+	 */
+	ereport(elevel,
+			(errmsg("\"%s\": %u pages, %d rows sampled, %.0f estimated total rows",
+					RelationGetRelationName(onerel),
+					onerel->rd_nblocks, numrows, *totalrows)));
+
 	return numrows;
 }
 
@@ -689,7 +714,7 @@ random_fract(void)
 	do
 	{
 		z = random();
-	} while (!(z > 0 && z < MAX_RANDOM_VALUE));
+	} while (z <= 0 || z >= MAX_RANDOM_VALUE);
 	return (double) z / (double) MAX_RANDOM_VALUE;
 }
 
@@ -852,6 +877,8 @@ compute_minimal_stats(VacAttrStats *stats,
 	double		total_width = 0;
 	bool		is_varlena = (!stats->attr->attbyval &&
 							  stats->attr->attlen == -1);
+	bool		is_varwidth = (!stats->attr->attbyval &&
+							   stats->attr->attlen < 0);
 	FmgrInfo	f_cmpeq;
 	typedef struct
 	{
@@ -897,7 +924,7 @@ compute_minimal_stats(VacAttrStats *stats,
 		nonnull_cnt++;
 
 		/*
-		 * If it's a varlena field, add up widths for average width
+		 * If it's a variable-width field, add up widths for average width
 		 * calculation.  Note that if the value is toasted, we use the
 		 * toasted width.  We don't bother with this calculation if it's a
 		 * fixed-width type.
@@ -920,6 +947,11 @@ compute_minimal_stats(VacAttrStats *stats,
 			}
 			value = PointerGetDatum(PG_DETOAST_DATUM(value));
 		}
+		else if (is_varwidth)
+		{
+			/* must be cstring */
+			total_width += strlen(DatumGetCString(value)) + 1;
+		}
 
 		/*
 		 * See if the value matches anything we're already tracking.
@@ -976,7 +1008,7 @@ compute_minimal_stats(VacAttrStats *stats,
 		stats->stats_valid = true;
 		/* Do the simple null-frac and width stats */
 		stats->stanullfrac = (double) null_cnt / (double) numrows;
-		if (is_varlena)
+		if (is_varwidth)
 			stats->stawidth = total_width / (double) nonnull_cnt;
 		else
 			stats->stawidth = stats->attrtype->typlen;
@@ -1027,11 +1059,15 @@ compute_minimal_stats(VacAttrStats *stats,
 			 */
 			int			f1 = nonnull_cnt - summultiple;
 			int			d = f1 + nmultiple;
-			double		numer, denom, stadistinct;
+			double		numer,
+						denom,
+						stadistinct;
+
+			numer = (double) numrows *(double) d;
 
-			numer = (double) numrows * (double) d;
 			denom = (double) (numrows - f1) +
-				(double) f1 * (double) numrows / totalrows;
+				(double) f1 *(double) numrows / totalrows;
+
 			stadistinct = numer / denom;
 			/* Clamp to sane range in case of roundoff error */
 			if (stadistinct < (double) d)
@@ -1099,8 +1135,8 @@ compute_minimal_stats(VacAttrStats *stats,
 			Datum	   *mcv_values;
 			float4	   *mcv_freqs;
 
-			/* Must copy the target values into TransactionCommandContext */
-			old_context = MemoryContextSwitchTo(TransactionCommandContext);
+			/* Must copy the target values into anl_context */
+			old_context = MemoryContextSwitchTo(anl_context);
 			mcv_values = (Datum *) palloc(num_mcv * sizeof(Datum));
 			mcv_freqs = (float4 *) palloc(num_mcv * sizeof(float4));
 			for (i = 0; i < num_mcv; i++)
@@ -1149,6 +1185,8 @@ compute_scalar_stats(VacAttrStats *stats,
 	double		total_width = 0;
 	bool		is_varlena = (!stats->attr->attbyval &&
 							  stats->attr->attlen == -1);
+	bool		is_varwidth = (!stats->attr->attbyval &&
+							   stats->attr->attlen < 0);
 	double		corr_xysum;
 	RegProcedure cmpFn;
 	SortFunctionKind cmpFnKind;
@@ -1188,7 +1226,7 @@ compute_scalar_stats(VacAttrStats *stats,
 		nonnull_cnt++;
 
 		/*
-		 * If it's a varlena field, add up widths for average width
+		 * If it's a variable-width field, add up widths for average width
 		 * calculation.  Note that if the value is toasted, we use the
 		 * toasted width.  We don't bother with this calculation if it's a
 		 * fixed-width type.
@@ -1211,6 +1249,11 @@ compute_scalar_stats(VacAttrStats *stats,
 			}
 			value = PointerGetDatum(PG_DETOAST_DATUM(value));
 		}
+		else if (is_varwidth)
+		{
+			/* must be cstring */
+			total_width += strlen(DatumGetCString(value)) + 1;
+		}
 
 		/* Add it to the list to be sorted */
 		values[values_cnt].value = value;
@@ -1303,7 +1346,7 @@ compute_scalar_stats(VacAttrStats *stats,
 		stats->stats_valid = true;
 		/* Do the simple null-frac and width stats */
 		stats->stanullfrac = (double) null_cnt / (double) numrows;
-		if (is_varlena)
+		if (is_varwidth)
 			stats->stawidth = total_width / (double) nonnull_cnt;
 		else
 			stats->stawidth = stats->attrtype->typlen;
@@ -1339,11 +1382,15 @@ compute_scalar_stats(VacAttrStats *stats,
 			 */
 			int			f1 = ndistinct - nmultiple + toowide_cnt;
 			int			d = f1 + nmultiple;
-			double		numer, denom, stadistinct;
+			double		numer,
+						denom,
+						stadistinct;
+
+			numer = (double) numrows *(double) d;
 
-			numer = (double) numrows * (double) d;
 			denom = (double) (numrows - f1) +
-				(double) f1 * (double) numrows / totalrows;
+				(double) f1 *(double) numrows / totalrows;
+
 			stadistinct = numer / denom;
 			/* Clamp to sane range in case of roundoff error */
 			if (stadistinct < (double) d)
@@ -1420,8 +1467,8 @@ compute_scalar_stats(VacAttrStats *stats,
 			Datum	   *mcv_values;
 			float4	   *mcv_freqs;
 
-			/* Must copy the target values into TransactionCommandContext */
-			old_context = MemoryContextSwitchTo(TransactionCommandContext);
+			/* Must copy the target values into anl_context */
+			old_context = MemoryContextSwitchTo(anl_context);
 			mcv_values = (Datum *) palloc(num_mcv * sizeof(Datum));
 			mcv_freqs = (float4 *) palloc(num_mcv * sizeof(float4));
 			for (i = 0; i < num_mcv; i++)
@@ -1506,8 +1553,8 @@ compute_scalar_stats(VacAttrStats *stats,
 				nvals = values_cnt;
 			Assert(nvals >= num_hist);
 
-			/* Must copy the target values into TransactionCommandContext */
-			old_context = MemoryContextSwitchTo(TransactionCommandContext);
+			/* Must copy the target values into anl_context */
+			old_context = MemoryContextSwitchTo(anl_context);
 			hist_values = (Datum *) palloc(num_hist * sizeof(Datum));
 			for (i = 0; i < num_hist; i++)
 			{
@@ -1535,8 +1582,8 @@ compute_scalar_stats(VacAttrStats *stats,
 			double		corr_xsum,
 						corr_x2sum;
 
-			/* Must copy the target values into TransactionCommandContext */
-			old_context = MemoryContextSwitchTo(TransactionCommandContext);
+			/* Must copy the target values into anl_context */
+			old_context = MemoryContextSwitchTo(anl_context);
 			corrs = (float4 *) palloc(sizeof(float4));
 			MemoryContextSwitchTo(old_context);
 
@@ -1636,6 +1683,15 @@ compare_mcvs(const void *a, const void *b)
  *		This could possibly be made to work, but it's not worth the trouble.
  *		Note analyze_rel() has seen to it that we won't come here when
  *		vacuuming pg_statistic itself.
+ *
+ *		Note: if two backends concurrently try to analyze the same relation,
+ *		the second one is likely to fail here with a "tuple concurrently
+ *		updated" error.  This is slightly annoying, but no real harm is done.
+ *		We could prevent the problem by using a stronger lock on the
+ *		relation for ANALYZE (ie, ShareUpdateExclusiveLock instead
+ *		of AccessShareLock); but that cure seems worse than the disease,
+ *		especially now that ANALYZE doesn't start a new transaction
+ *		for each relation.	The lock could be held for a long time...
  */
 static void
 update_attstats(Oid relid, int natts, VacAttrStats **vacattrstats)
@@ -1643,17 +1699,11 @@ update_attstats(Oid relid, int natts, VacAttrStats **vacattrstats)
 	Relation	sd;
 	int			attno;
 
-	/*
-	 * We use an ExclusiveLock on pg_statistic to ensure that only one
-	 * backend is writing it at a time --- without that, we might have to
-	 * deal with concurrent updates here, and it's not worth the trouble.
-	 */
-	sd = heap_openr(StatisticRelationName, ExclusiveLock);
+	sd = heap_openr(StatisticRelationName, RowExclusiveLock);
 
 	for (attno = 0; attno < natts; attno++)
 	{
 		VacAttrStats *stats = vacattrstats[attno];
-		FmgrInfo	out_function;
 		HeapTuple	stup,
 					oldtup;
 		int			i,
@@ -1662,14 +1712,11 @@ update_attstats(Oid relid, int natts, VacAttrStats **vacattrstats)
 		Datum		values[Natts_pg_statistic];
 		char		nulls[Natts_pg_statistic];
 		char		replaces[Natts_pg_statistic];
-		Relation	irelations[Num_pg_statistic_indices];
 
 		/* Ignore attr if we weren't able to collect stats */
 		if (!stats->stats_valid)
 			continue;
 
-		fmgr_info(stats->attrtype->typoutput, &out_function);
-
 		/*
 		 * Construct a new pg_statistic tuple
 		 */
@@ -1706,7 +1753,8 @@ update_attstats(Oid relid, int natts, VacAttrStats **vacattrstats)
 					numdatums[n] = Float4GetDatum(stats->stanumbers[k][n]);
 				/* XXX knows more than it should about type float4: */
 				arry = construct_array(numdatums, nnum,
-									   false, sizeof(float4), 'i');
+									   FLOAT4OID,
+									   sizeof(float4), false, 'i');
 				values[i++] = PointerGetDatum(arry);	/* stanumbersN */
 			}
 			else
@@ -1717,32 +1765,16 @@ update_attstats(Oid relid, int natts, VacAttrStats **vacattrstats)
 		}
 		for (k = 0; k < STATISTIC_NUM_SLOTS; k++)
 		{
-			int			ntxt = stats->numvalues[k];
-
-			if (ntxt > 0)
+			if (stats->numvalues[k] > 0)
 			{
-				Datum	   *txtdatums = (Datum *) palloc(ntxt * sizeof(Datum));
 				ArrayType  *arry;
 
-				for (n = 0; n < ntxt; n++)
-				{
-					/*
-					 * Convert data values to a text string to be inserted
-					 * into the text array.
-					 */
-					Datum		stringdatum;
-
-					stringdatum =
-						FunctionCall3(&out_function,
-									  stats->stavalues[k][n],
-							  ObjectIdGetDatum(stats->attrtype->typelem),
-								  Int32GetDatum(stats->attr->atttypmod));
-					txtdatums[n] = DirectFunctionCall1(textin, stringdatum);
-					pfree(DatumGetPointer(stringdatum));
-				}
-				/* XXX knows more than it should about type text: */
-				arry = construct_array(txtdatums, ntxt,
-									   false, -1, 'i');
+				arry = construct_array(stats->stavalues[k],
+									   stats->numvalues[k],
+									   stats->attr->atttypid,
+									   stats->attrtype->typlen,
+									   stats->attrtype->typbyval,
+									   stats->attrtype->typalign);
 				values[i++] = PointerGetDatum(arry);	/* stavaluesN */
 			}
 			else
@@ -1773,18 +1805,14 @@ update_attstats(Oid relid, int natts, VacAttrStats **vacattrstats)
 		{
 			/* No, insert new tuple */
 			stup = heap_formtuple(sd->rd_att, values, nulls);
-			heap_insert(sd, stup);
+			simple_heap_insert(sd, stup);
 		}
 
-		/* update indices too */
-		CatalogOpenIndices(Num_pg_statistic_indices, Name_pg_statistic_indices,
-						   irelations);
-		CatalogIndexInsert(irelations, Num_pg_statistic_indices, sd, stup);
-		CatalogCloseIndices(Num_pg_statistic_indices, irelations);
+		/* update indexes too */
+		CatalogUpdateIndexes(sd, stup);
 
 		heap_freetuple(stup);
 	}
 
-	/* close rel, but hold lock till upcoming commit */
-	heap_close(sd, NoLock);
+	heap_close(sd, RowExclusiveLock);
 }
