@@ -3,40 +3,25 @@
  * int8.c
  *	  Internal 64-bit integer operations
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/int8.c,v 1.42 2002/09/18 21:35:22 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/int8.c,v 1.48 2003/09/25 06:58:04 petere Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
 #include <ctype.h>
-#include <time.h>
 #include <math.h>
-#include <float.h>
-#include <limits.h>
 
+#include "libpq/pqformat.h"
 #include "utils/int8.h"
 
 
 #define MAXINT8LEN		25
 
-#ifndef INT_MAX
-#define INT_MAX (0x7FFFFFFFL)
-#endif
-#ifndef INT_MIN
-#define INT_MIN (-INT_MAX-1)
-#endif
-#ifndef SHRT_MAX
-#define SHRT_MAX (0x7FFF)
-#endif
-#ifndef SHRT_MIN
-#define SHRT_MIN (-SHRT_MAX-1)
-#endif
-
 
 /***********************************************************************
  **
@@ -51,7 +36,7 @@
 /*
  * scanint8 --- try to parse a string into an int8.
  *
- * If errorOK is false, elog a useful error message if the string is bad.
+ * If errorOK is false, ereport a useful error message if the string is bad.
  * If errorOK is true, just return "false" for bad input.
  */
 bool
@@ -98,7 +83,9 @@ scanint8(const char *str, bool errorOK, int64 *result)
 		if (errorOK)
 			return false;
 		else
-			elog(ERROR, "Bad int8 external representation \"%s\"", str);
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
+				  errmsg("invalid input syntax for type bigint: \"%s\"", str)));
 	}
 
 	/* process digits */
@@ -111,7 +98,9 @@ scanint8(const char *str, bool errorOK, int64 *result)
 			if (errorOK)
 				return false;
 			else
-				elog(ERROR, "int8 value out of range: \"%s\"", str);
+				ereport(ERROR,
+						(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+						 errmsg("integer out of range")));
 		}
 		tmp = newtmp;
 	}
@@ -122,7 +111,9 @@ scanint8(const char *str, bool errorOK, int64 *result)
 		if (errorOK)
 			return false;
 		else
-			elog(ERROR, "Bad int8 external representation \"%s\"", str);
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
+				  errmsg("invalid input syntax for type bigint: \"%s\"", str)));
 	}
 
 	*result = (sign < 0) ? -tmp : tmp;
@@ -154,12 +145,37 @@ int8out(PG_FUNCTION_ARGS)
 	char		buf[MAXINT8LEN + 1];
 
 	if ((len = snprintf(buf, MAXINT8LEN, INT64_FORMAT, val)) < 0)
-		elog(ERROR, "Unable to format int8");
+		elog(ERROR, "could not format int8");
 
 	result = pstrdup(buf);
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		int8recv			- converts external binary format to int8
+ */
+Datum
+int8recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+
+	PG_RETURN_INT64(pq_getmsgint64(buf));
+}
+
+/*
+ *		int8send			- converts int8 to binary format
+ */
+Datum
+int8send(PG_FUNCTION_ARGS)
+{
+	int64		arg1 = PG_GETARG_INT64(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+	pq_sendint64(&buf, arg1);
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
 
 /*----------------------------------------------------------
  *	Relational operators for int8s, including cross-data-type comparisons.
@@ -504,6 +520,11 @@ int8div(PG_FUNCTION_ARGS)
 	int64		val1 = PG_GETARG_INT64(0);
 	int64		val2 = PG_GETARG_INT64(1);
 
+	if (val2 == 0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DIVISION_BY_ZERO),
+				 errmsg("division by zero")));
+
 	PG_RETURN_INT64(val1 / val2);
 }
 
@@ -528,6 +549,11 @@ int8mod(PG_FUNCTION_ARGS)
 	int64		val2 = PG_GETARG_INT64(1);
 	int64		result;
 
+	if (val2 == 0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DIVISION_BY_ZERO),
+				 errmsg("division by zero")));
+
 	result = val1 / val2;
 	result *= val2;
 	result = val1 - result;
@@ -621,6 +647,11 @@ int84div(PG_FUNCTION_ARGS)
 	int64		val1 = PG_GETARG_INT64(0);
 	int32		val2 = PG_GETARG_INT32(1);
 
+	if (val2 == 0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DIVISION_BY_ZERO),
+				 errmsg("division by zero")));
+
 	PG_RETURN_INT64(val1 / val2);
 }
 
@@ -657,6 +688,11 @@ int48div(PG_FUNCTION_ARGS)
 	int32		val1 = PG_GETARG_INT32(0);
 	int64		val2 = PG_GETARG_INT64(1);
 
+	if (val2 == 0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DIVISION_BY_ZERO),
+				 errmsg("division by zero")));
+
 	PG_RETURN_INT64(val1 / val2);
 }
 
@@ -745,7 +781,9 @@ int84(PG_FUNCTION_ARGS)
 
 	/* Test for overflow by reverse-conversion. */
 	if ((int64) result != val)
-		elog(ERROR, "int8 conversion to int4 is out of range");
+		ereport(ERROR,
+				(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+				 errmsg("integer out of range")));
 
 	PG_RETURN_INT32(result);
 }
@@ -768,7 +806,9 @@ int82(PG_FUNCTION_ARGS)
 
 	/* Test for overflow by reverse-conversion. */
 	if ((int64) result != val)
-		elog(ERROR, "int8 conversion to int2 is out of range");
+		ereport(ERROR,
+				(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+				 errmsg("integer out of range")));
 
 	PG_RETURN_INT16(result);
 }
@@ -804,7 +844,9 @@ dtoi8(PG_FUNCTION_ARGS)
 	result = (int64) val;
 
 	if ((float8) result != val)
-		elog(ERROR, "Floating point conversion to int8 is out of range");
+		ereport(ERROR,
+				(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+				 errmsg("integer out of range")));
 
 	PG_RETURN_INT64(result);
 }
@@ -841,7 +883,9 @@ ftoi8(PG_FUNCTION_ARGS)
 	result = (int64) dval;
 
 	if ((float8) result != dval)
-		elog(ERROR, "Floating point conversion to int8 is out of range");
+		ereport(ERROR,
+				(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+				 errmsg("integer out of range")));
 
 	PG_RETURN_INT64(result);
 }
@@ -856,7 +900,9 @@ i8tooid(PG_FUNCTION_ARGS)
 
 	/* Test for overflow by reverse-conversion. */
 	if ((int64) result != val)
-		elog(ERROR, "int8 conversion to OID is out of range");
+		ereport(ERROR,
+				(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+				 errmsg("OID out of range")));
 
 	PG_RETURN_OID(result);
 }
