@@ -33,7 +33,7 @@
  *	  ENHANCEMENTS, OR MODIFICATIONS.
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/pl/plperl/plperl.c,v 1.35 2002/09/21 18:39:26 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/pl/plperl/plperl.c,v 1.40 2003/09/04 15:16:39 tgl Exp $
  *
  **********************************************************************/
 
@@ -47,7 +47,6 @@
 /* postgreSQL stuff */
 #include "executor/spi.h"
 #include "commands/trigger.h"
-#include "utils/elog.h"
 #include "fmgr.h"
 #include "access/heapam.h"
 #include "tcop/tcopprot.h"
@@ -92,8 +91,6 @@ typedef struct plperl_proc_desc
  * Global data
  **********************************************************************/
 static int	plperl_firstcall = 1;
-static int	plperl_call_level = 0;
-static int	plperl_restart_in_progress = 0;
 static PerlInterpreter *plperl_interp = NULL;
 static HV  *plperl_proc_hash = NULL;
 
@@ -104,6 +101,7 @@ static void plperl_init_all(void);
 static void plperl_init_interp(void);
 
 Datum		plperl_call_handler(PG_FUNCTION_ARGS);
+void		plperl_init(void);
 
 static Datum plperl_func_handler(PG_FUNCTION_ARGS);
 
@@ -131,29 +129,21 @@ perm_fmgr_info(Oid functionId, FmgrInfo *finfo)
 }
 
 /**********************************************************************
- * plperl_init_all()		- Initialize all
+ * plperl_init()			- Initialize everything that can be
+ *							  safely initialized during postmaster
+ *							  startup.
+ *
+ * DO NOT make this static --- it has to be callable by preload
  **********************************************************************/
-static void
-plperl_init_all(void)
+void
+plperl_init(void)
 {
-
 	/************************************************************
 	 * Do initialization only once
 	 ************************************************************/
 	if (!plperl_firstcall)
 		return;
 
-
-	/************************************************************
-	 * Destroy the existing Perl interpreter
-	 ************************************************************/
-	if (plperl_interp != NULL)
-	{
-		perl_destruct(plperl_interp);
-		perl_free(plperl_interp);
-		plperl_interp = NULL;
-	}
-
 	/************************************************************
 	 * Free the proc hash table
 	 ************************************************************/
@@ -164,6 +154,16 @@ plperl_init_all(void)
 		plperl_proc_hash = NULL;
 	}
 
+	/************************************************************
+	 * Destroy the existing Perl interpreter
+	 ************************************************************/
+	if (plperl_interp != NULL)
+	{
+		perl_destruct(plperl_interp);
+		perl_free(plperl_interp);
+		plperl_interp = NULL;
+	}
+
 	/************************************************************
 	 * Now recreate a new Perl interpreter
 	 ************************************************************/
@@ -172,6 +172,26 @@ plperl_init_all(void)
 	plperl_firstcall = 0;
 }
 
+/**********************************************************************
+ * plperl_init_all()		- Initialize all
+ **********************************************************************/
+static void
+plperl_init_all(void)
+{
+
+	/************************************************************
+	 * Execute postmaster-startup safe initialization
+	 ************************************************************/
+	if (plperl_firstcall)
+		plperl_init();
+
+	/************************************************************
+	 * Any other initialization that must be done each time a new
+	 * backend starts -- currently none
+	 ************************************************************/
+
+}
+
 
 /**********************************************************************
  * plperl_init_interp() - Create the Perl interpreter
@@ -196,14 +216,12 @@ plperl_init_interp(void)
 
 	plperl_interp = perl_alloc();
 	if (!plperl_interp)
-		elog(ERROR, "plperl_init_interp(): could not allocate perl interpreter");
+		elog(ERROR, "could not allocate perl interpreter");
 
 	perl_construct(plperl_interp);
 	perl_parse(plperl_interp, plperl_init_shared_libs, 3, embedding, NULL);
 	perl_run(plperl_interp);
 
-
-
 	/************************************************************
 	 * Initialize the proc and query hash tables
 	 ************************************************************/
@@ -212,7 +230,6 @@ plperl_init_interp(void)
 }
 
 
-
 /**********************************************************************
  * plperl_call_handler		- This is the only visible function
  *				  of the PL interpreter. The PostgreSQL
@@ -229,20 +246,15 @@ plperl_call_handler(PG_FUNCTION_ARGS)
 	Datum		retval;
 
 	/************************************************************
-	 * Initialize interpreters on first call
+	 * Initialize interpreter
 	 ************************************************************/
-	if (plperl_firstcall)
-		plperl_init_all();
+	plperl_init_all();
 
 	/************************************************************
 	 * Connect to SPI manager
 	 ************************************************************/
 	if (SPI_connect() != SPI_OK_CONNECT)
-		elog(ERROR, "plperl: cannot connect to SPI manager");
-	/************************************************************
-	 * Keep track about the nesting of Perl-SPI-Perl-... calls
-	 ************************************************************/
-	plperl_call_level++;
+		elog(ERROR, "could not connect to SPI manager");
 
 	/************************************************************
 	 * Determine if called as function or trigger and
@@ -250,7 +262,9 @@ plperl_call_handler(PG_FUNCTION_ARGS)
 	 ************************************************************/
 	if (CALLED_AS_TRIGGER(fcinfo))
 	{
-		elog(ERROR, "plperl: can't use perl in triggers yet.");
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("cannot use perl in triggers yet")));
 
 		/*
 		 * retval = PointerGetDatum(plperl_trigger_handler(fcinfo));
@@ -261,8 +275,6 @@ plperl_call_handler(PG_FUNCTION_ARGS)
 	else
 		retval = plperl_func_handler(fcinfo);
 
-	plperl_call_level--;
-
 	return retval;
 }
 
@@ -272,13 +284,11 @@ plperl_call_handler(PG_FUNCTION_ARGS)
  *		create the anonymous subroutine whose text is in the SV.
  *		Returns the SV containing the RV to the closure.
  **********************************************************************/
-static
-SV *
+static SV  *
 plperl_create_sub(char *s, bool trusted)
 {
 	dSP;
-
-	SV		   *subref = NULL;
+	SV		   *subref;
 	int			count;
 
 	ENTER;
@@ -286,10 +296,24 @@ plperl_create_sub(char *s, bool trusted)
 	PUSHMARK(SP);
 	XPUSHs(sv_2mortal(newSVpv(s, 0)));
 	PUTBACK;
+
+	/*
+	 * G_KEEPERR seems to be needed here, else we don't recognize compile
+	 * errors properly.  Perhaps it's because there's another level of
+	 * eval inside mksafefunc?
+	 */
 	count = perl_call_pv((trusted ? "mksafefunc" : "mkunsafefunc"),
 						 G_SCALAR | G_EVAL | G_KEEPERR);
 	SPAGAIN;
 
+	if (count != 1)
+	{
+		PUTBACK;
+		FREETMPS;
+		LEAVE;
+		elog(ERROR, "didn't get a return item from mksafefunc");
+	}
+
 	if (SvTRUE(ERRSV))
 	{
 		POPs;
@@ -299,9 +323,6 @@ plperl_create_sub(char *s, bool trusted)
 		elog(ERROR, "creation of function failed: %s", SvPV(ERRSV, PL_na));
 	}
 
-	if (count != 1)
-		elog(ERROR, "creation of function failed - no return from mksafefunc");
-
 	/*
 	 * need to make a deep copy of the return. it comes off the stack as a
 	 * temporary.
@@ -318,12 +339,13 @@ plperl_create_sub(char *s, bool trusted)
 		 * subref is our responsibility because it is not mortal
 		 */
 		SvREFCNT_dec(subref);
-		elog(ERROR, "plperl_create_sub: didn't get a code ref");
+		elog(ERROR, "didn't get a code ref");
 	}
 
 	PUTBACK;
 	FREETMPS;
 	LEAVE;
+
 	return subref;
 }
 
@@ -352,21 +374,18 @@ plperl_init_shared_libs(pTHX)
  * plperl_call_perl_func()		- calls a perl function through the RV
  *			stored in the prodesc structure. massages the input parms properly
  **********************************************************************/
-static
-SV *
+static SV  *
 plperl_call_perl_func(plperl_proc_desc * desc, FunctionCallInfo fcinfo)
 {
 	dSP;
-
 	SV		   *retval;
 	int			i;
 	int			count;
 
-
 	ENTER;
 	SAVETMPS;
 
-	PUSHMARK(sp);
+	PUSHMARK(SP);
 	for (i = 0; i < desc->nargs; i++)
 	{
 		if (desc->arg_is_rel[i])
@@ -401,7 +420,9 @@ plperl_call_perl_func(plperl_proc_desc * desc, FunctionCallInfo fcinfo)
 		}
 	}
 	PUTBACK;
-	count = perl_call_sv(desc->reference, G_SCALAR | G_EVAL | G_KEEPERR);
+
+	/* Do NOT use G_KEEPERR here */
+	count = perl_call_sv(desc->reference, G_SCALAR | G_EVAL);
 
 	SPAGAIN;
 
@@ -410,7 +431,7 @@ plperl_call_perl_func(plperl_proc_desc * desc, FunctionCallInfo fcinfo)
 		PUTBACK;
 		FREETMPS;
 		LEAVE;
-		elog(ERROR, "plperl: didn't get a return item from function");
+		elog(ERROR, "didn't get a return item from function");
 	}
 
 	if (SvTRUE(ERRSV))
@@ -419,21 +440,19 @@ plperl_call_perl_func(plperl_proc_desc * desc, FunctionCallInfo fcinfo)
 		PUTBACK;
 		FREETMPS;
 		LEAVE;
-		elog(ERROR, "plperl: error from function: %s", SvPV(ERRSV, PL_na));
+		elog(ERROR, "error from function: %s", SvPV(ERRSV, PL_na));
 	}
 
 	retval = newSVsv(POPs);
 
-
 	PUTBACK;
 	FREETMPS;
 	LEAVE;
 
 	return retval;
-
-
 }
 
+
 /**********************************************************************
  * plperl_func_handler()		- Handler for regular function calls
  **********************************************************************/
@@ -443,23 +462,10 @@ plperl_func_handler(PG_FUNCTION_ARGS)
 	plperl_proc_desc *prodesc;
 	SV		   *perlret;
 	Datum		retval;
-	sigjmp_buf	save_restart;
 
 	/* Find or compile the function */
 	prodesc = compile_plperl_function(fcinfo->flinfo->fn_oid, false);
 
-	/* Set up error handling */
-	memcpy(&save_restart, &Warn_restart, sizeof(save_restart));
-
-	if (sigsetjmp(Warn_restart, 1) != 0)
-	{
-		memcpy(&Warn_restart, &save_restart, sizeof(Warn_restart));
-		plperl_restart_in_progress = 1;
-		if (--plperl_call_level == 0)
-			plperl_restart_in_progress = 0;
-		siglongjmp(Warn_restart, 1);
-	}
-
 	/************************************************************
 	 * Call the Perl function
 	 ************************************************************/
@@ -472,7 +478,7 @@ plperl_func_handler(PG_FUNCTION_ARGS)
 	 * because SPI_finish would free it).
 	 ************************************************************/
 	if (SPI_finish() != SPI_OK_FINISH)
-		elog(ERROR, "plperl: SPI_finish() failed");
+		elog(ERROR, "SPI_finish() failed");
 
 	if (!(perlret && SvOK(perlret)))
 	{
@@ -490,14 +496,6 @@ plperl_func_handler(PG_FUNCTION_ARGS)
 
 	SvREFCNT_dec(perlret);
 
-	memcpy(&Warn_restart, &save_restart, sizeof(Warn_restart));
-	if (plperl_restart_in_progress)
-	{
-		if (--plperl_call_level == 0)
-			plperl_restart_in_progress = 0;
-		siglongjmp(Warn_restart, 1);
-	}
-
 	return retval;
 }
 
@@ -520,7 +518,7 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 							 ObjectIdGetDatum(fn_oid),
 							 0, 0, 0);
 	if (!HeapTupleIsValid(procTup))
-		elog(ERROR, "plperl: cache lookup for proc %u failed", fn_oid);
+		elog(ERROR, "cache lookup failed for function %u", fn_oid);
 	procStruct = (Form_pg_proc) GETSTRUCT(procTup);
 
 	/************************************************************
@@ -578,7 +576,9 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 		 ************************************************************/
 		prodesc = (plperl_proc_desc *) malloc(sizeof(plperl_proc_desc));
 		if (prodesc == NULL)
-			elog(ERROR, "plperl: out of memory");
+			ereport(ERROR,
+					(errcode(ERRCODE_OUT_OF_MEMORY),
+					 errmsg("out of memory")));
 		MemSet(prodesc, 0, sizeof(plperl_proc_desc));
 		prodesc->proname = strdup(internal_proname);
 		prodesc->fn_xmin = HeapTupleHeaderGetXmin(procTup->t_data);
@@ -594,7 +594,7 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 		{
 			free(prodesc->proname);
 			free(prodesc);
-			elog(ERROR, "plperl: cache lookup for language %u failed",
+			elog(ERROR, "cache lookup failed for language %u",
 				 procStruct->prolang);
 		}
 		langStruct = (Form_pg_language) GETSTRUCT(langTup);
@@ -614,7 +614,7 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 			{
 				free(prodesc->proname);
 				free(prodesc);
-				elog(ERROR, "plperl: cache lookup for return type %u failed",
+				elog(ERROR, "cache lookup failed for type %u",
 					 procStruct->prorettype);
 			}
 			typeStruct = (Form_pg_type) GETSTRUCT(typeTup);
@@ -628,16 +628,18 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 				{
 					free(prodesc->proname);
 					free(prodesc);
-					elog(ERROR, "plperl functions cannot return type %s"
-						 "\n\texcept when used as triggers",
-						 format_type_be(procStruct->prorettype));
+					ereport(ERROR,
+							(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+							 errmsg("trigger functions may only be called as triggers")));
 				}
 				else
 				{
 					free(prodesc->proname);
 					free(prodesc);
-					elog(ERROR, "plperl functions cannot return type %s",
-						 format_type_be(procStruct->prorettype));
+					ereport(ERROR,
+							(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						 errmsg("plperl functions cannot return type %s",
+								format_type_be(procStruct->prorettype))));
 				}
 			}
 
@@ -645,7 +647,9 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 			{
 				free(prodesc->proname);
 				free(prodesc);
-				elog(ERROR, "plperl: return types of tuples not supported yet");
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				   errmsg("plperl functions cannot return tuples yet")));
 			}
 
 			perm_fmgr_info(typeStruct->typinput, &(prodesc->result_in_func));
@@ -670,7 +674,7 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 				{
 					free(prodesc->proname);
 					free(prodesc);
-					elog(ERROR, "plperl: cache lookup for argument type %u failed",
+					elog(ERROR, "cache lookup failed for type %u",
 						 procStruct->proargtypes[i]);
 				}
 				typeStruct = (Form_pg_type) GETSTRUCT(typeTup);
@@ -680,8 +684,10 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 				{
 					free(prodesc->proname);
 					free(prodesc);
-					elog(ERROR, "plperl functions cannot take type %s",
-						 format_type_be(procStruct->proargtypes[i]));
+					ereport(ERROR,
+							(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						   errmsg("plperl functions cannot take type %s",
+						   format_type_be(procStruct->proargtypes[i]))));
 				}
 
 				if (typeStruct->typrelid != InvalidOid)
@@ -713,7 +719,7 @@ compile_plperl_function(Oid fn_oid, bool is_trigger)
 		{
 			free(prodesc->proname);
 			free(prodesc);
-			elog(ERROR, "plperl: cannot create internal procedure %s",
+			elog(ERROR, "could not create internal procedure \"%s\"",
 				 internal_proname);
 		}
 
@@ -751,6 +757,10 @@ plperl_build_tuple_argument(HeapTuple tuple, TupleDesc tupdesc)
 
 	for (i = 0; i < tupdesc->natts; i++)
 	{
+		/* ignore dropped attributes */
+		if (tupdesc->attrs[i]->attisdropped)
+			continue;
+
 		/************************************************************
 		 * Get the attribute name
 		 ************************************************************/
@@ -778,8 +788,8 @@ plperl_build_tuple_argument(HeapTuple tuple, TupleDesc tupdesc)
 						   ObjectIdGetDatum(tupdesc->attrs[i]->atttypid),
 								 0, 0, 0);
 		if (!HeapTupleIsValid(typeTup))
-			elog(ERROR, "plperl: Cache lookup for attribute '%s' type %u failed",
-				 attname, tupdesc->attrs[i]->atttypid);
+			elog(ERROR, "cache lookup failed for type %u",
+				 tupdesc->attrs[i]->atttypid);
 
 		typoutput = ((Form_pg_type) GETSTRUCT(typeTup))->typoutput;
 		typelem = ((Form_pg_type) GETSTRUCT(typeTup))->typelem;
